#include "StdAfx.h"
#include "AudioMonitor.h"

CommandReader* AudioMonitor::myReader = NULL;
AudioQueue* AudioMonitor::queue = NULL;
FFT* AudioMonitor::myFFT = NULL;
bool AudioMonitor::quit = false;
int AudioMonitor::threshold = 6;
HWAVEIN AudioMonitor::wavdev = NULL;
HANDLE AudioMonitor::myThread = NULL;

AudioMonitor::AudioMonitor(CommandReader* in, int device)
{
	if( singleton == false ){
		singleton = true;
		myReader = in;
		quit = false;
		clean = false;
		//Wave input stuff
		WAVEFORMATEX wavForm = 
		{
			WAVE_FORMAT_PCM, 
			NUMCHANNELS, 
			SAMPLESPERSEC,
			AVGBYTESPERSEC,
			BLOCKALIGN,
			BITSPERSAMPLE,
			0
		};

		myFFT = new FFT( BINS );
		if( myFFT == NULL ){
			delete this;
		}

		if( waveInOpen( &wavdev, device, &wavForm, (int)&AudioMonitor::WaveCallback, 0, CALLBACK_FUNCTION )!=0 ){
			delete this;
		}

		queue = new AudioQueue( wavdev );
		if( queue == NULL ){
			delete this;
		}
		waveInStart( wavdev );
		myThread = CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE)&AudioMonitor::ThreadFunc, NULL, 0, NULL );
		if( myThread == NULL ){
			delete this;
		}
	}
}

AudioMonitor::AudioMonitor( HWND win, CString myDevice )
{
	if( singleton == false ){
		singleton = true;
		myReader = new CommandReader( win );
		quit = false;
		clean = true;
		int dev = -1;
		int numDevs = waveInGetNumDevs();
		for( int x =0; x<numDevs; x++ ){
			CString tmpName;
			WAVEINCAPS device;

			waveInGetDevCaps( x, &device, sizeof( device ) );
			tmpName = device.szPname;
			if( _tcscmp( myDevice, device.szPname ) == 0)
			{
				dev = x;
			}
		}
		//Wave input stuff
		WAVEFORMATEX wavForm = 
		{
			WAVE_FORMAT_PCM, 
			NUMCHANNELS, 
			SAMPLESPERSEC,
			AVGBYTESPERSEC,
			BLOCKALIGN,
			BITSPERSAMPLE,
			0
		};

		myFFT = new FFT( BINS );
		if( myFFT == NULL ){
			delete this;
		}

		if( waveInOpen( &wavdev, dev, &wavForm, (int)&AudioMonitor::WaveCallback, 0, CALLBACK_FUNCTION )!=0 ){
			delete this;
		}

		queue = new AudioQueue( wavdev );
		if( queue == NULL ){
			delete this;
		}
		waveInStart( wavdev );
		myThread = CreateThread( NULL, 0, (LPTHREAD_START_ROUTINE)&AudioMonitor::ThreadFunc, NULL, 0, NULL );
		if( myThread == NULL ){
			delete this;
		}
	}
}
void AudioMonitor::WaveCallback( HWAVEIN waveIn, UINT Msg, DWORD_PTR Callback, DWORD_PTR Param1, DWORD_PTR Param2 )
{
	if( Msg == MM_WIM_DATA ){
		queue->Put( (LPWAVEHDR)Param1 );
	}
}

AudioMonitor::~AudioMonitor(void)
{
	singleton = false;
	if( !quit ){
		quit = true;
		while( myThread != NULL ){
			Sleep( 1000 );
		}
	}

	if( myReader != NULL ){
		if( clean )
		{
			delete myReader;
		}
		myReader = NULL;
	}

	if( queue != NULL ){
		delete queue;
		queue = NULL;
	}

	waveInClose( wavdev );
	wavdev = NULL;
	if( myFFT != NULL ){
		delete myFFT;
		myFFT = NULL;
	}
}

void AudioMonitor::Quit(){
	quit = true;
}

void AudioMonitor::ThreadFunc()
{
	WORD aBuffer[ BUFFERSIZE ];
	char Last = '\0';
	char LastSent = '\0';

	while( !quit ){
		DWORD tickCount;
		while( queue->Get( aBuffer, &tickCount ) ){
			int lBuffer[BINS/2];
			myFFT->Run( (short*)aBuffer, lBuffer );
			int ToneH = -1;
			int ToneL = -1;
			int CountH = 0;
			int CountL = 0;
			int AvgH = 0;
			int AvgL = 0;
			int Avg = 0;
			int ThresholdAGC = 0;

			for( int n = BIN0-1; n<=(BIN7+1); n++ ){
				Avg += lBuffer[ n ];
			}

			for( int x = 0; x<4; x++ ){
				Avg -= lBuffer[DTMFHigh[ x ]];
				Avg -= lBuffer[DTMFLow[ x ]];
				AvgH += lBuffer[DTMFHigh[ x ]];
				AvgL += lBuffer[DTMFLow[ x ]];
			}

			Avg = (Avg)/(BIN7-BIN0 -5);
			ThresholdAGC = Avg * threshold / 2;
			AvgH = AvgH * 3 /4;
			AvgL = AvgL * 3 /4;

			for( int y = 0; y<4; y++ ){
				if( lBuffer[DTMFHigh[y]] > ThresholdAGC ){
					ToneH = y;
					CountH++;
				}
				if( lBuffer[DTMFLow[y]] > ThresholdAGC ){
					ToneL = y;
					CountL++;
				}
			}
			if( (CountH == 1)&&(CountL == 1) ){
				char c = ToneDecode[ToneL][ToneH];
				if( c == Last ){
					if( LastSent!=c ){
						myReader->ProcessChar( c );
						LastSent = c;
					}
				}
				Last = c;
			}
			else{
				Last = '\0';
				LastSent = '\0';
			}
		}
	}
	myThread = NULL;
}