#include "StdAfx.h"
#include "AudioQueue.h"

AudioQueue::AudioQueue(HWAVEIN in)
{
	InitializeCriticalSection( &lock );
	myWav = in;
	head = 0;
	tail = 0;
	count = 0;
	for( int n = 0; n<NUMBUFFERS; n++ ){
		queue[n].header.lpData = NULL;
		queue[n].prepared = false;
		queue[n].bufferReady = false;
		index[n]=-1;
	}
	for( int x = 0; x<NUMBUFFERS; x++ ){
		queue[x].header.lpData= (LPSTR)malloc( BUFFERSIZE );
		if( queue[x].header.lpData == NULL ){
			delete this;
		}
		queue[x].header.dwBufferLength = BUFFERSIZE;
		queue[x].header.dwFlags = 0;
		queue[x].header.dwUser = (DWORD)x;
		int sample = waveInPrepareHeader( in, &queue[x].header, sizeof(queue[x].header));
		if( sample!=0 ){
			delete this;
		}
		queue[x].prepared = true;
		waveInAddBuffer( in, &queue[x].header, sizeof( queue[x].header ) );
	}
}

AudioQueue::~AudioQueue(void)
{
	for( int n = 0; n<NUMBUFFERS; n++ ){
		if( queue[n].prepared ){
			waveInUnprepareHeader( myWav, &queue[n].header, sizeof( queue[n].header ) );
			queue[n].prepared = false;
		}

		if( queue[n].header.lpData != NULL ){
			delete queue[n].header.lpData;
			queue[n].header.lpData = NULL;
		}
	}
	DeleteCriticalSection( &lock );

}

bool AudioQueue::Put( LPWAVEHDR inHeader ){
	int Next = (head + 1) % NUMBUFFERS;
	int ind = (int)inHeader->dwUser;

	if( Next == tail ){
		return false;
	}
	EnterCriticalSection( &lock );
	index[ head ] = ind;
	queue[ ind ].bufferReady = true;
	queue[ ind ].tickCount = GetTickCount();
	head = Next;
	count++;
	LeaveCriticalSection( &lock );
	return true;
}

bool AudioQueue::Get( PWORD data, PDWORD tickCount ){

	int ind;

	if( head == tail ){
		return false;
	}
	EnterCriticalSection( &lock );
	memcpy( data, queue[tail].header.lpData, BUFFERSIZE );
	if( tickCount != NULL){
		*tickCount = queue[ tail ].tickCount;
	}

	ind = tail;
	queue[ ind ].bufferReady = false;
	tail = (tail+1) % NUMBUFFERS;
	count--;
	LeaveCriticalSection( &lock );

	waveInAddBuffer( myWav, &queue[ind].header, sizeof(queue[ind].header) );
	
	return true;
}

int AudioQueue::GetCount(){
	EnterCriticalSection( &lock );
	int ret = count;
	LeaveCriticalSection( &lock );
	return ret;
}