#include "StdAfx.h"
#include "FFT.h"
#include <math.h>

#define PI 3.1415926535

FFT::FFT(int length)
{
	bins = length;
	if( !InitializeReverse() ){
		delete this;
		return;
	}
	if( !InitializeTrig() ){
		delete this;
		return;
	}
	if( !InitializeWindow() ){
		delete this;
		return;
	}
}

FFT::~FFT(void)
{
	/*if(window != NULL){
		free( window );
		window = NULL;
	}
	if( sins != NULL ){
		free( sins );
		sins = NULL;
	}
	if( coss != NULL ){
		free( coss );
		coss = NULL;
	}
	if( reversed != NULL ){
		free( reversed );
		reversed = NULL;
	}*/
}

/**
 * Initializes an array of shorts, probably with the 
 * values being the bitwise reverse of the index.
 */
bool FFT::InitializeReverse(){
	reversed = (short*)malloc( bins*sizeof(short));
	if( reversed==NULL ){
		return false;
	}
	for( int x = 0; x<bins; x++ ){
		int i = 0;
		for( int y = bins/4; y>0; y>>=1 ){
			i=(i>>1) + (x & y ? bins/2 : 0);
		}
		reversed[x] = (short)i;
	}
	return true;
}

bool FFT::InitializeTrig(){
	sins = (short*)malloc( bins*sizeof(short) );
	coss = (short*)malloc( bins*sizeof(short) );
	if( (sins == NULL)&& (coss == NULL) ){
		return false;
	}
	for( int x = 0; x<bins/2; x++ ){

		double dSin = floor(-32768.0*sin(2*PI*x/bins) + .5);
		double dCos = floor(-32768.0*cos(2*PI*x/bins) + .5);

		if( dSin>32767.5 ){
			dSin = 32767.0;
		}
		if( dCos>32767.5 ){
			dCos = 32767.0;
		}
		
		sins[ reversed[x]/2 ] = (short)dSin;
		coss[ reversed[x]/2 ] = (short)dCos;
	}

	return true;
}

bool FFT::InitializeWindow(){
	window = (short*)malloc( bins*sizeof(short) );
	if( window == NULL ){
		return false;
	}

	for( int x = 0; x<bins; x++ ){
		double hamming = 0.54-0.46*cos(2*PI*x/bins);
		window[x] = (short)floor( hamming*32767+0.5 );
	}
	return true;
}

void FFT::Run( short *BufferIn, int *BufferOut ){

	short* end = BufferIn+bins;
	short* index;
	short* curWindow = window;
	
	inputClipped = false;

	for( index = BufferIn; index<end; index++ ){
		short val = *index;
		if( val>29490 || val<-29490 ){
			inputClipped = true;
		}
		*index=(short)(((int)val*(int)(*curWindow))>>15);
		curWindow++;
	}

	short* alpha;
	short* beta;
	
	int Group = bins/4;
	
	while( Group > 0 ){
		alpha = BufferIn;
		beta = BufferIn + Group * 2;
		int mark = 0;
		while( alpha < end ){
			
			short mySin = sins[ mark ];
			short myCos = coss[ mark ];
			short* oEnd = beta;

			while( alpha < oEnd ){
				int V = ((int)beta[0] * myCos + (int)beta[1] * mySin) >>15;
				int W = ((int)beta[0] * mySin - (int)beta[1] * myCos) >>15;

				beta[0] = (short)((alpha[0] + V) >> 1 );
				alpha[0] = (short)(beta[0] - V);

				beta[1] = (short)((alpha[1] - W) >> 1 );
				alpha[1] = (short)(beta[1] + W);

				alpha+=2;
				beta+=2;
			}
			alpha = beta;
			beta+= Group*2;
			mark++;
		}
		Group>>=1;
	}
	
	alpha = reversed + 1;
	beta = reversed + bins / 2 - 1;

	while( alpha<=beta ){
		short mySin = sins[ alpha[0]/2 ];
		short myCos = coss[ alpha[0]/2 ];

		short* A = BufferIn + *alpha;
		short* B = BufferIn + *beta;

		int Rminus = A[0] - B[0];
		int Rplus = Rminus + (B[0]<<1);

		int Iminus = A[1] - B[1];
		int Iplus = Iminus + (B[1]<<1);

		int S = ((int)mySin * Rminus - (int)myCos*Iplus) >>15;
		int T = ((int)myCos * Rminus + (int)mySin*Iplus) >>15;

		A[0] = (short)((Rplus + S) >>1);
		B[0] = (short)(A[0] - S);

		A[1] = (short)((Iminus + T) >>1);
		B[1] = (short)(A[1] - Iminus);

		alpha++;
		beta--;
	}
	
	
	BufferIn[0] += BufferIn[1];
	BufferIn[1] = 0;

	int* Out = BufferOut;
	short* rev = reversed;

	long Sum = 0;
	max = 0;

	for(int n = 0; n<bins/2; n++){
		int R = BufferIn[rev[0]];
		int I = BufferIn[rev[0]+1];
		int root;
		int mask;
		int A2 = (R*R) + (I*I);

		if( A2<0 ){
			A2=0;
		}

		if( A2 > 4194304L ){
			root = 32;

			do{
				mask = A2/root;
				root = (root+mask) >> 1;
			}while( labs(root - mask) > 1 );
			root*=16;
		}else{
			root = 512;
			A2*=256;

			do{
				mask = A2/root;
				root = (root+mask) >> 1;
			}while( labs(root-mask) >1 );
		}
		*BufferOut++=root;
		Sum += root;

		if( root>max ){
			max = root;
		}
		rev++;
	}
	average = (int)(Sum/(bins/2));

}